@@ -40,6 +40,8 @@ sub new {
     relay_log_info      => undef,
     mysqlbinlog_version => undef,
     dir                 => undef,
+    client_bindir       => undef,
+    client_libdir       => undef,
     prefix              => undef,
     cur_log             => undef,
     end_log             => undef,
@@ -47,6 +49,11 @@ sub new {
     debug               => undef,
     @_,
   };
+  $self->{mysqlbinlog} = MHA::NodeUtil::client_cli_prefix(
+    "mysqlbinlog",
+    $self->{client_bindir},
+    $self->{client_libdir},
+  );
   return bless $self, $class;
 }
 
@@ -55,8 +62,8 @@ sub get_apply_arg($) {
   my $arg  = "";
   if (
     should_suppress_row_format(
-      $self->{mysqlbinlog_version},
-      $self->{mysql_version}
+      $self->{mysqlbinlog_version}, $self->{mysql_version},
+      $self->{mysqlbinlog}
     )
     )
   {
@@ -83,37 +90,43 @@ sub mysqlbinlog_version_ge {
 
 sub die_if_too_old_version {
   my $mysqlbinlog_version = shift;
+  my $mysqlbinlog         = shift;
   if ( !mysqlbinlog_ge_50($mysqlbinlog_version) ) {
     croak
-"mysqlbinlog version is $mysqlbinlog_version. This is too old. MHA supports MySQL version 5.0 (mysqlbinlog version 3.2) or higher. Recommended mysqlbinlog version is 3.3+, which is included in MySQL 5.1 or higher.\n";
+"$mysqlbinlog version is $mysqlbinlog_version. This is too old. MHA supports MySQL version 5.0 (mysqlbinlog version 3.2) or higher. Recommended mysqlbinlog version is 3.3+, which is included in MySQL 5.1 or higher.\n";
   }
 }
 
 sub init_mysqlbinlog($) {
   my $self = shift;
   eval {
-    unless ( $self->{mysql_version} )
-    {
+    unless ( $self->{mysql_version} ) {
       croak "mysql version not found.\n";
     }
-    my $v = `mysqlbinlog --version`;
+    my $v = `$self->{mysqlbinlog} --version`;
+    my ( $high, $low ) = MHA::NodeUtil::system_rc($?);
+    if ( $high || $low ) {
+      croak "$self->{mysqlbinlog} version command failed with rc $high:$low, "
+        . "please verify PATH, LD_LIBRARY_PATH, and client options\n";
+    }
     chomp($v);
     if ( $v =~ /Ver (\d+\.\d+)/ ) {
       $self->{mysqlbinlog_version} = $1;
     }
-    croak "mysqlbinlog version not found!\n"
+    croak "$self->{mysqlbinlog} version not found!\n"
       unless ( $self->{mysqlbinlog_version} );
-    die_if_too_old_version( $self->{mysqlbinlog_version} );
+    die_if_too_old_version( $self->{mysqlbinlog_version},
+      $self->{mysqlbinlog} );
     if ( !mysqlbinlog_ge_51( $self->{mysqlbinlog_version} ) ) {
       print
-"mysqlbinlog version is $self->{mysqlbinlog_version} (included in MySQL Client 5.0 or lower). This is not recommended. Consider upgrading MySQL Client to 5.1 or higher.\n";
+"$self->{mysqlbinlog} version is $self->{mysqlbinlog_version} (included in MySQL Client 5.0 or lower). This is not recommended. Consider upgrading MySQL Client to 5.1 or higher.\n";
     }
     if ( !mysqlbinlog_ge_51( $self->{mysqlbinlog_version} )
       && $self->{mysql_version}
       && MHA::NodeUtil::mysql_version_ge( $self->{mysql_version}, "5.1.0" ) )
     {
       croak sprintf(
-"mysqlbinlog is %s (included in MySQL Client 5.0 or lower), but MySQL server version is %s. mysqlbinlog can not parse row based events. Terminating script for safety reasons.\n",
+"$self->{mysqlbinlog} is %s (included in MySQL Client 5.0 or lower), but MySQL server version is %s. mysqlbinlog can not parse row based events. Terminating script for safety reasons.\n",
         $self->{mysqlbinlog_version},
         $self->{mysql_version}
       );
@@ -214,7 +227,7 @@ sub get_relaydir_and_filename_from_rinfo($$) {
 
   unless ( $current_relay_log_file =~ m/^\// ) {
     $current_relay_log_file =~ s/^\.\///;
-    $datadir                =~ s/\/$//;
+    $datadir =~ s/\/$//;
     $current_relay_log_file = $datadir . "/" . $current_relay_log_file;
     $relay_dir              = $datadir;
   }
@@ -326,12 +339,13 @@ sub get_post_file($) {
 sub should_suppress_row_format {
   my $mysqlbinlog_version = shift;
   my $mysql_version       = shift;
+  my $mysqlbinlog         = shift;
   my $suppress_row_format = 0;
   if ( mysqlbinlog_ge_51($mysqlbinlog_version)
     && !MHA::NodeUtil::mysql_version_ge( $mysql_version, "5.1.0" ) )
   {
     print
-"mysqlbinlog is 5.1 or higher, and MySQL version on the target server is 5.0 or lower. So using mysqlbinlog --base64-output=never to disable BINLOG events..\n";
+"$mysqlbinlog is 5.1 or higher, and MySQL version on the target server is 5.0 or lower. So using mysqlbinlog --base64-output=never to disable BINLOG events..\n";
     $suppress_row_format = 1;
   }
   return $suppress_row_format;
@@ -348,10 +362,11 @@ sub get_relaydir_and_files_from_rinfo {
 }
 
 sub is_binlog_head_readable($) {
+  my $self = shift;
   my $file = shift;
 
   # higher than binlog file header (4 bytes)
-  return system("mysqlbinlog --stop-position=5 $file > /dev/null");
+  return system("$self->{mysqlbinlog} --stop-position=5 $file > /dev/null");
 }
 
 sub get_end_binlog_fde($$$) {
@@ -495,12 +510,22 @@ sub dump_mysqlbinlog($$$$$$) {
   my $disable_log_bin     = $self->{disable_log_bin};
 
   my $rc;
-
+  my $binlog_file = "$binlog_dir/$from_file";
+  my $file_size   = -s $binlog_file;
+  if ( $from_pos > 4 && $from_pos > $file_size ) {
+    print
+"Target file $binlog_file size=$file_size, this is smaller than exec pos $from_pos. Skipping.\n";
+    return 1;
+  }
+  elsif ( $from_pos > 4 && $from_pos == $file_size ) {
+    print "No additional binlog events found.\n";
+    return 1;
+  }
   my $command =
     "echo \"# Binary/Relay log file $from_file started\" >> $out_diff_file";
   system($command);
 
-  $command = "mysqlbinlog --start-position=$from_pos ";
+  $command = "$self->{mysqlbinlog} --start-position=$from_pos ";
   if ($suppress_row_format) {
     $command .= " --base64-output=never";
   }
@@ -508,14 +533,20 @@ sub dump_mysqlbinlog($$$$$$) {
     $command .= " --disable-log-bin";
   }
 
-  $command .= " $binlog_dir/$from_file";
-  $command .= "| filter_mysqlbinlog";
+  $command .= " $binlog_file";
+  if ( !$self->{skip_filter} ) {
+    $command .= "| filter_mysqlbinlog";
+  }
   $command .= " >> $out_diff_file";
+  if ( $self->{debug} ) {
+    print "Executing command: $command\n";
+  }
+
   $rc = system($command);
   if ($rc) {
     my ( $high, $low ) = MHA::NodeUtil::system_rc($rc);
     croak
-"FATAL: mysqlbinlog to binlog/relaylog file $from_file, generating to $out_diff_file failed with rc $high:$low!\n";
+"FATAL: $self->{mysqlbinlog} to binlog/relaylog file $from_file, generating to $out_diff_file failed with rc $high:$low!\n";
   }
   return 0;
 }
@@ -538,7 +569,8 @@ sub concat_all_binlogs_from($$$$) {
   my $suppress_row_format;
   if ( !$handle_raw_binlog ) {
     $suppress_row_format =
-      should_suppress_row_format( $mysqlbinlog_version, $mysql_version );
+      should_suppress_row_format( $mysqlbinlog_version, $mysql_version,
+      $self->{mysqlbinlog} );
   }
   print
 " Concat binary/relay logs from $binlog_prefix.$start_num pos $start_pos to $binlog_prefix.$end_num EOF into $outfile ..\n";
@@ -562,13 +594,17 @@ sub concat_all_binlogs_from($$$$) {
         $first_file );
     }
     else {
-      $self->dump_mysqlbinlog( $self->{dir}, $from_file, $from_pos, $outfile,
-        $suppress_row_format );
+      my $rc =
+        $self->dump_mysqlbinlog( $self->{dir}, $from_file, $from_pos,
+        $outfile, $suppress_row_format );
+      if ($rc) {
+        return $rc;
+      }
     }
   }
   if ( -f $outfile ) {
     if ($handle_raw_binlog) {
-      croak "$outfile is broken!\n" if is_binlog_head_readable($outfile);
+      croak "$outfile is broken!\n" if $self->is_binlog_head_readable($outfile);
       if ( -s $outfile <=
         $self->get_end_binlog_fde( dirname($outfile), basename($outfile) ) )
       {
@@ -616,7 +652,7 @@ sub concat_generated_binlogs($$$) {
           $out );
       }
     }
-    if ( is_binlog_head_readable($out) ) {
+    if ( $self->is_binlog_head_readable($out) ) {
       croak "$out is broken!\n";
     }
   }
