@@ -26,6 +26,7 @@ use Carp qw(croak);
 use MHA::NodeConst;
 use File::Path;
 use Errno();
+use Socket qw(NI_NUMERICHOST getaddrinfo getnameinfo);
 
 sub create_dir_if($) {
   my $dir = shift;
@@ -80,8 +81,8 @@ sub file_copy {
   my $log_output  = shift;
   my $ssh_port    = shift;
   $ssh_port = 22 unless ($ssh_port);
-
-  my $ssh_user_host = $ssh_user . '@' . $ssh_host;
+  my $dsn_host = $ssh_host =~ m{:} ? '[' . $ssh_host . ']' : $ssh_host;
+  my $ssh_user_host = $ssh_user . '@' . $dsn_host;
   my ( $from, $to );
   if ($to_remote) {
     $from = $local_file;
@@ -150,13 +151,23 @@ sub drop_file_if($) {
 
 sub get_ip {
   my $host = shift;
-  my ( $bin_addr_host, $addr_host );
+  my ( $err, @bin_addr_host, $addr_host );
   if ( defined($host) ) {
-    $bin_addr_host = gethostbyname($host);
-    unless ($bin_addr_host) {
-      croak "Failed to get IP address on host $host!\n";
-    }
-    $addr_host = sprintf( "%vd", $bin_addr_host );
+    ( $err, @bin_addr_host ) = getaddrinfo($host);
+    croak "Failed to get IP address on host $host: $err\n" if $err;
+
+    # We take the first ip address that is returned by getaddrinfo
+    ( $err, $addr_host ) = getnameinfo($bin_addr_host[0]->{addr}, NI_NUMERICHOST);
+    croak "Failed to convert IP address for host $host: $err\n" if $err;
+
+    # for IPv6 (and it works with IPv4 and hostnames as well):
+    # - DBD-MySQL expects [::] format
+    # - scp requires [::] format
+    # - ssh requires :: format
+    # - mysql tools require :: format
+    # The code in MHA is expected to use [] when it is running scp
+    # when it connects with DBD-MySQL
+
     return $addr_host;
   }
   return;
@@ -260,4 +271,33 @@ sub escape_for_mysql_command {
   return $ret;
 }
 
+sub client_cli_prefix {
+  my ( $exe, $bindir, $libdir ) = @_;
+  croak "unexpected client binary $exe\n" unless $exe =~ /^mysql(?:binlog)?$/;
+  my %env = ( LD_LIBRARY_PATH => $libdir );
+  my $cli = $exe;
+  if ($bindir) {
+    if ( ref $bindir eq "ARRAY" ) {
+      $env{'PATH'} = $bindir;
+    }
+    elsif ( ref $bindir eq "" ) {
+      $cli = escape_for_shell("$bindir/$exe");
+    }
+  }
+  for my $k ( keys %env ) {
+    if ( my $v = $env{$k} ) {
+      my @dirs = ref $v eq "ARRAY" ? @{$v} : ( ref $v eq "" ? ($v) : () );
+      @dirs = grep { $_ && !/:/ } @dirs;
+      if (@dirs) {
+        $cli = "$k="
+          . join( ":", ( map { escape_for_shell($_) } @dirs ), "\$$k" )
+          . " $cli";
+      }
+    }
+  }
+
+  # $cli .= " --no-defaults";
+  return $cli;
+}
+
 1;
